"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const os_1 = __importDefault(require("os"));
const util_1 = require("util");
const NodeFS_1 = require("./NodeFS");
const path_1 = require("./path");
var FakeFS_1 = require("./FakeFS");
exports.normalizeLineEndings = FakeFS_1.normalizeLineEndings;
var path_2 = require("./path");
exports.PortablePath = path_2.PortablePath;
exports.Filename = path_2.Filename;
var path_3 = require("./path");
exports.npath = path_3.npath;
exports.ppath = path_3.ppath;
exports.toFilename = path_3.toFilename;
var AliasFS_1 = require("./AliasFS");
exports.AliasFS = AliasFS_1.AliasFS;
var FakeFS_2 = require("./FakeFS");
exports.FakeFS = FakeFS_2.FakeFS;
var CwdFS_1 = require("./CwdFS");
exports.CwdFS = CwdFS_1.CwdFS;
var JailFS_1 = require("./JailFS");
exports.JailFS = JailFS_1.JailFS;
var LazyFS_1 = require("./LazyFS");
exports.LazyFS = LazyFS_1.LazyFS;
var NoFS_1 = require("./NoFS");
exports.NoFS = NoFS_1.NoFS;
var NodeFS_2 = require("./NodeFS");
exports.NodeFS = NodeFS_2.NodeFS;
var PosixFS_1 = require("./PosixFS");
exports.PosixFS = PosixFS_1.PosixFS;
var ProxiedFS_1 = require("./ProxiedFS");
exports.ProxiedFS = ProxiedFS_1.ProxiedFS;
var VirtualFS_1 = require("./VirtualFS");
exports.VirtualFS = VirtualFS_1.VirtualFS;
var ZipFS_1 = require("./ZipFS");
exports.ZipFS = ZipFS_1.ZipFS;
var ZipOpenFS_1 = require("./ZipOpenFS");
exports.ZipOpenFS = ZipOpenFS_1.ZipOpenFS;
function getTempName(prefix) {
    const tmpdir = path_1.npath.toPortablePath(os_1.default.tmpdir());
    const hash = Math.ceil(Math.random() * 0x100000000).toString(16).padStart(8, `0`);
    return path_1.ppath.join(tmpdir, `${prefix}${hash}`);
}
function patchFs(patchedFs, fakeFs) {
    const SYNC_IMPLEMENTATIONS = new Set([
        `accessSync`,
        `appendFileSync`,
        `createReadStream`,
        `chmodSync`,
        `closeSync`,
        `copyFileSync`,
        `lstatSync`,
        `mkdirSync`,
        `openSync`,
        `readSync`,
        `readlinkSync`,
        `readFileSync`,
        `readdirSync`,
        `readlinkSync`,
        `realpathSync`,
        `renameSync`,
        `rmdirSync`,
        `statSync`,
        `symlinkSync`,
        `unlinkSync`,
        `utimesSync`,
        `watch`,
        `writeFileSync`,
        `writeSync`,
    ]);
    const ASYNC_IMPLEMENTATIONS = new Set([
        `accessPromise`,
        `appendFilePromise`,
        `chmodPromise`,
        `closePromise`,
        `copyFilePromise`,
        `lstatPromise`,
        `mkdirPromise`,
        `openPromise`,
        `readdirPromise`,
        `realpathPromise`,
        `readFilePromise`,
        `readdirPromise`,
        `readlinkPromise`,
        `renamePromise`,
        `rmdirPromise`,
        `statPromise`,
        `symlinkPromise`,
        `unlinkPromise`,
        `utimesPromise`,
        `writeFilePromise`,
        `writeSync`,
    ]);
    const setupFn = (target, name, replacement) => {
        const orig = target[name];
        if (typeof orig === `undefined`)
            return;
        target[name] = replacement;
        if (typeof orig[util_1.promisify.custom] !== `undefined`) {
            replacement[util_1.promisify.custom] = orig[util_1.promisify.custom];
        }
    };
    setupFn(patchedFs, `existsSync`, (p) => {
        try {
            return fakeFs.existsSync(p);
        }
        catch (error) {
            return false;
        }
    });
    setupFn(patchedFs, `exists`, (p, ...args) => {
        const hasCallback = typeof args[args.length - 1] === `function`;
        const callback = hasCallback ? args.pop() : () => { };
        process.nextTick(() => {
            fakeFs.existsPromise(p).then(exists => {
                callback(exists);
            }, () => {
                callback(false);
            });
        });
    });
    setupFn(patchedFs, `read`, (p, buffer, ...args) => {
        const hasCallback = typeof args[args.length - 1] === `function`;
        const callback = hasCallback ? args.pop() : () => { };
        process.nextTick(() => {
            fakeFs.readPromise(p, buffer, ...args).then(bytesRead => {
                callback(null, bytesRead, buffer);
            }, error => {
                callback(error);
            });
        });
    });
    for (const fnName of ASYNC_IMPLEMENTATIONS) {
        const fakeImpl = fakeFs[fnName].bind(fakeFs);
        const origName = fnName.replace(/Promise$/, ``);
        setupFn(patchedFs, origName, (...args) => {
            const hasCallback = typeof args[args.length - 1] === `function`;
            const callback = hasCallback ? args.pop() : () => { };
            process.nextTick(() => {
                fakeImpl(...args).then((result) => {
                    callback(null, result);
                }, (error) => {
                    callback(error);
                });
            });
        });
    }
    for (const fnName of SYNC_IMPLEMENTATIONS) {
        const fakeImpl = fakeFs[fnName].bind(fakeFs);
        const origName = fnName;
        setupFn(patchedFs, origName, fakeImpl);
    }
    patchedFs.realpathSync.native = patchedFs.realpathSync;
    patchedFs.realpath.native = patchedFs.realpath;
}
exports.patchFs = patchFs;
function extendFs(realFs, fakeFs) {
    const patchedFs = Object.create(realFs);
    patchFs(patchedFs, fakeFs);
    return patchedFs;
}
exports.extendFs = extendFs;
const tmpdirs = new Set();
let cleanExitRegistered = false;
function registerCleanExit() {
    if (!cleanExitRegistered)
        cleanExitRegistered = true;
    else
        return;
    const cleanExit = () => {
        process.off(`exit`, cleanExit);
        for (const p of tmpdirs) {
            tmpdirs.delete(p);
            try {
                exports.xfs.removeSync(p);
            }
            catch (_a) {
                // Too bad if there's an error
            }
        }
    };
    process.on(`exit`, cleanExit);
}
exports.xfs = Object.assign(new NodeFS_1.NodeFS(), {
    detachTemp(p) {
        tmpdirs.delete(p);
    },
    mktempSync(cb) {
        registerCleanExit();
        while (true) {
            const p = getTempName(`xfs-`);
            try {
                this.mkdirSync(p);
            }
            catch (error) {
                if (error.code === `EEXIST`) {
                    continue;
                }
                else {
                    throw error;
                }
            }
            const realP = this.realpathSync(p);
            tmpdirs.add(realP);
            if (typeof cb !== `undefined`) {
                try {
                    return cb(realP);
                }
                finally {
                    if (tmpdirs.has(realP)) {
                        tmpdirs.delete(realP);
                        try {
                            this.removeSync(realP);
                        }
                        catch (_a) {
                            // Too bad if there's an error
                        }
                    }
                }
            }
            else {
                return p;
            }
        }
    },
    async mktempPromise(cb) {
        registerCleanExit();
        while (true) {
            const p = getTempName(`xfs-`);
            try {
                await this.mkdirPromise(p);
            }
            catch (error) {
                if (error.code === `EEXIST`) {
                    continue;
                }
                else {
                    throw error;
                }
            }
            const realP = await this.realpathPromise(p);
            tmpdirs.add(realP);
            if (typeof cb !== `undefined`) {
                try {
                    return await cb(realP);
                }
                finally {
                    if (tmpdirs.has(realP)) {
                        tmpdirs.delete(realP);
                        try {
                            await this.removePromise(realP);
                        }
                        catch (_a) {
                            // Too bad if there's an error
                        }
                    }
                }
            }
            else {
                return realP;
            }
        }
    },
});
