"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const fslib_1 = require("@yarnpkg/fslib");
const chalk_1 = __importDefault(require("chalk"));
const comment_json_1 = __importDefault(require("comment-json"));
const dynamicRequire_1 = require("./dynamicRequire");
const TEMPLATE = (relPnpApiPath, module, { setupEnv = false, usePnpify = false }) => [
    `#!/usr/bin/env node\n`,
    `\n`,
    `const {createRequire, createRequireFromPath} = require(\`module\`);\n`,
    `const {resolve} = require(\`path\`);\n`,
    `\n`,
    `const relPnpApiPath = ${JSON.stringify(fslib_1.npath.fromPortablePath(relPnpApiPath))};\n`,
    `\n`,
    `const absPnpApiPath = resolve(__dirname, relPnpApiPath);\n`,
    `const absRequire = (createRequire || createRequireFromPath)(absPnpApiPath);\n`,
    `\n`,
    `// Setup the environment to be able to require ${module}\n`,
    `require(absPnpApiPath).setup();\n`,
    ...(setupEnv || usePnpify ? [
        `\n`,
        `// Prepare the environment (to be ready in case of child_process.spawn etc)\n`,
    ] : []),
    ...(setupEnv ? [
        `process.env.NODE_OPTIONS = process.env.NODE_OPTIONS || \`\`;\n`,
        `process.env.NODE_OPTIONS += \` -r \${absPnpApiPath}\`;\n`,
    ] : []),
    ...(usePnpify ? [
        `process.env.NODE_OPTIONS += \` -r \${pnpifyResolution}\`;\n`,
        `\n`,
        `// Apply PnPify to the current process\n`,
        `absRequire(\`@yarnpkg/pnpify\`).patchFs();\n`,
    ] : []),
    `\n`,
    `// Defer to the real ${module} your application uses\n`,
    `module.exports = absRequire(\`${module}\`);\n`,
].join(``);
const addVSCodeWorkspaceSettings = async (pnpApi, settings) => {
    const topLevelInformation = pnpApi.getPackageInformation(pnpApi.topLevel);
    const projectRoot = fslib_1.npath.toPortablePath(topLevelInformation.packageLocation);
    const settingsPath = fslib_1.ppath.join(projectRoot, `.vscode/settings.json`);
    const content = await fslib_1.xfs.existsPromise(settingsPath)
        ? await fslib_1.xfs.readFilePromise(settingsPath, `utf8`)
        : `{}`;
    const data = comment_json_1.default.parse(content);
    const patched = `${comment_json_1.default.stringify(Object.assign(Object.assign({}, data), settings), null, 2)}\n`;
    await fslib_1.xfs.mkdirpPromise(fslib_1.ppath.dirname(settingsPath));
    await fslib_1.xfs.changeFilePromise(settingsPath, patched, {
        automaticNewlines: true,
    });
};
class Wrapper {
    constructor(name, { pnpApi, target }) {
        this.paths = new Map();
        this.name = name;
        this.pnpApi = pnpApi;
        this.target = target;
    }
    async writeManifest() {
        const absWrapperPath = fslib_1.ppath.join(this.target, this.name, `package.json`);
        const topLevelInformation = this.pnpApi.getPackageInformation(this.pnpApi.topLevel);
        const dependencyReference = topLevelInformation.packageDependencies.get(this.name);
        const pkgInformation = this.pnpApi.getPackageInformation(this.pnpApi.getLocator(this.name, dependencyReference));
        if (pkgInformation === null)
            throw new Error(`Assertion failed: Package ${this.name} isn't a dependency of the top-level`);
        const manifest = dynamicRequire_1.dynamicRequire(fslib_1.npath.join(pkgInformation.packageLocation, `package.json`));
        await fslib_1.xfs.mkdirpPromise(fslib_1.ppath.dirname(absWrapperPath));
        await fslib_1.xfs.writeFilePromise(absWrapperPath, JSON.stringify({
            name: this.name,
            version: `${manifest.version}-pnpify`,
            main: manifest.main,
            type: `commonjs`,
        }, null, 2));
    }
    async writeBinary(relPackagePath, options = {}) {
        const absPackagePath = await this.writeFile(relPackagePath, options);
        await fslib_1.xfs.chmodPromise(absPackagePath, 0o755);
    }
    async writeFile(relPackagePath, options = {}) {
        const topLevelInformation = this.pnpApi.getPackageInformation(this.pnpApi.topLevel);
        const projectRoot = fslib_1.npath.toPortablePath(topLevelInformation.packageLocation);
        const absWrapperPath = fslib_1.ppath.join(this.target, this.name, relPackagePath);
        const relProjectPath = fslib_1.ppath.relative(projectRoot, absWrapperPath);
        const absPnpApiPath = fslib_1.npath.toPortablePath(this.pnpApi.resolveRequest(`pnpapi`, null));
        const relPnpApiPath = fslib_1.ppath.relative(fslib_1.ppath.dirname(absWrapperPath), absPnpApiPath);
        await fslib_1.xfs.mkdirpPromise(fslib_1.ppath.dirname(absWrapperPath));
        await fslib_1.xfs.writeFilePromise(absWrapperPath, TEMPLATE(relPnpApiPath, fslib_1.ppath.join(this.name, relPackagePath), options));
        this.paths.set(relPackagePath, relProjectPath);
        return absWrapperPath;
    }
    getProjectPathTo(relPackagePath) {
        const relProjectPath = this.paths.get(relPackagePath);
        if (typeof relProjectPath === `undefined`)
            throw new Error(`Assertion failed: Expected path to have been registered`);
        return relProjectPath;
    }
}
const generateTypescriptWrapper = async (pnpApi, target) => {
    const wrapper = new Wrapper(`typescript`, { pnpApi, target });
    await wrapper.writeManifest();
    await wrapper.writeBinary(`bin/tsc`);
    await wrapper.writeBinary(`bin/tsserver`);
    await wrapper.writeFile(`lib/tsc.js`);
    await wrapper.writeFile(`lib/tsserver.js`);
    await wrapper.writeFile(`lib/typescript.js`);
    await addVSCodeWorkspaceSettings(pnpApi, {
        [`typescript.tsdk`]: fslib_1.npath.fromPortablePath(fslib_1.ppath.dirname(wrapper.getProjectPathTo(`lib/tsserver.js`))),
    });
};
exports.generateEslintWrapper = async (pnpApi, target) => {
    const wrapper = new Wrapper(`eslint`, { pnpApi, target });
    await wrapper.writeManifest();
    await wrapper.writeBinary(`bin/eslint.js`);
    await wrapper.writeFile(`lib/api.js`);
    await addVSCodeWorkspaceSettings(pnpApi, {
        [`eslint.nodePath`]: fslib_1.npath.fromPortablePath(fslib_1.ppath.dirname(fslib_1.ppath.dirname(fslib_1.ppath.dirname(wrapper.getProjectPathTo(`lib/api.js`))))),
    });
};
exports.generatePrettierWrapper = async (pnpApi, target) => {
    const wrapper = new Wrapper(`prettier`, { pnpApi, target });
    await wrapper.writeManifest();
    await wrapper.writeBinary(`index.js`);
    await addVSCodeWorkspaceSettings(pnpApi, {
        [`prettier.prettierPath`]: fslib_1.npath.fromPortablePath(wrapper.getProjectPathTo(`index.js`)),
    });
};
const SDKS = [
    [`eslint`, exports.generateEslintWrapper],
    [`prettier`, exports.generatePrettierWrapper],
    [`typescript`, generateTypescriptWrapper],
];
exports.generateSdk = async (pnpApi) => {
    const topLevelInformation = pnpApi.getPackageInformation(pnpApi.topLevel);
    const projectRoot = fslib_1.npath.toPortablePath(topLevelInformation.packageLocation);
    const targetFolder = fslib_1.ppath.join(projectRoot, `.vscode/pnpify`);
    if (fslib_1.xfs.existsSync(targetFolder)) {
        console.log(`Cleaning up the existing SDK files...`);
        await fslib_1.xfs.removePromise(targetFolder);
    }
    console.log(`Installing fresh SDKs for ${chalk_1.default.magenta(projectRoot)}:`);
    console.log(``);
    let skippedSome = false;
    for (const [pkgName, generateWrapper] of SDKS) {
        const displayName = pkgName.replace(/^[a-z]/g, $0 => $0.toUpperCase());
        if (topLevelInformation.packageDependencies.has(pkgName)) {
            console.log(`  ${chalk_1.default.green(`✓`)} ${displayName}`);
            await generateWrapper(pnpApi, targetFolder);
        }
        else {
            console.log(`  ${chalk_1.default.yellow(`•`)} ${displayName} (dependency not found; skipped)`);
            skippedSome = true;
        }
    }
    if (skippedSome) {
        console.log(``);
        console.log(`Note that in order to be detected those packages have to be listed as top-level`);
        console.log(`dependencies (listing them into each individual workspace won't be enough).`);
    }
};
